import logging
import threading

from typing import List, cast, Dict, Any
from typing_extensions import TypeAlias

from eportal import utils
from eportal.utils import BatchProcessor, process_int
from eportal.patchsource import PatchSource
from config import ANOMALIES_QUEUE_TIMEOUT as QUEUE_TIMEOUT, ANOMALIES_QUEUE_MAX_BATCH_SIZE as QUEUE_MAX_BATCH_SIZE

CheckinT: TypeAlias = Dict[str, Any]

logger = logging.getLogger(__name__)


ANOMALY_DETECTOR_VERSION = 1
SEND_TIMEOUT = 30.0
INT_FIELDS = [
    'uptime',
    'patch_level',
    'ts',
    'proxy',
    'last_stop',
    'ltimestamp',
    'last',
]


def process_kernel_anomaly(data: List[CheckinT]) -> None:
    for checkin in data:
        if detect_kernel_anomaly(checkin):
            kernel_anomaly_consumer.append(checkin)


def detect_kernel_anomaly(checkin: CheckinT) -> bool:
    reason = checkin.get('reason')
    uptime = checkin.get('uptime')
    patch_level = checkin.get('patch_level')
    try:
        state_ts = int(checkin['state']['ts'])
    except (KeyError, TypeError, ValueError):
        state_ts = None
    last_stop = checkin.get('last_stop')
    timestamp = checkin.get('ts')
    fields = [reason, uptime, patch_level, state_ts, last_stop, timestamp]
    first_update_after_reboot_marker = False
    creash_soon_after_update_marker = False
    no_proper_shutdown_marker = False
    if all(fields):
        reason = cast(str, reason)
        uptime = cast(int, uptime)
        patch_level = cast(int, patch_level)
        state_ts = cast(int, state_ts)
        last_stop = cast(int, last_stop)
        timestamp = cast(int, timestamp)
        if uptime < 300 and patch_level == -1:
            first_update_after_reboot_marker = True
        if (timestamp - uptime) > state_ts > (timestamp - uptime - 1800) and reason == 'update':
            creash_soon_after_update_marker = True
        if last_stop < state_ts:
            no_proper_shutdown_marker = True
    markers = [first_update_after_reboot_marker, creash_soon_after_update_marker, no_proper_shutdown_marker]

    if all(markers):
        logger.info('Kernel anomaly detected: %s', checkin)
        return True
    return False


CheckinBatchProcessor = BatchProcessor[CheckinT]


@CheckinBatchProcessor(timeout=QUEUE_TIMEOUT, batch_size=QUEUE_MAX_BATCH_SIZE)
def kernel_anomaly_consumer(batch: List[CheckinT]) -> None:
    send_anomalies_to_patchserver(batch)


def fields_to_int(checkin: CheckinT) -> CheckinT:
    for field in INT_FIELDS:
        process_int(checkin, field, default=None)

    return checkin


def send_anomalies_to_patchserver(batch: List[CheckinT]) -> None:
    batch = [fields_to_int(checkin) for checkin in batch]
    ps = PatchSource()
    url = ps.get_url('/eportal-api/anomalies')
    if not ps.is_configured():
        return
    logging.debug('Sending %i anomalies to patchserver', len(batch))

    payload = {
        'anomalies': batch,
        'detector_version': ANOMALY_DETECTOR_VERSION,
    }
    try:
        with utils.get_session() as s:
            response = s.post(url, json=payload, timeout=SEND_TIMEOUT, **utils.get_req_args(auth=ps.auth))
            response.raise_for_status()
        logger.info('%s anomalies sent to patchserver.', len(batch))
    except Exception as e:
        logger.error('Failed to send anomalies to patchserver Url: %s, error: %s', url, e)


def anomaly_queue_processor(stop: threading.Event) -> threading.Thread:
    return kernel_anomaly_consumer.start(stop, 'Kernel anomaly consumer error')
